/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

/******************************** C H E C K **************************************/

int se_after_select = 0;

static Node s_dummy_node, *dummy_node = &s_dummy_node;

Node *check_SelectCaseStmt (root)
Node *root;
{
	Typ *typ = root->SelectCaseStmt.Expr->com.typ;
	int type = typ->etype->type;
	int rank = RANK_FROM_SHAPE (typ->shape);
	Node *cac;
	struct S_CaseConstruct *scac;

/* error checking */

	if (type != TYP_INTEGER && type != TYP_CHARACTER && type != TYP_LOGICAL) {
		sprintf (serr, "expression in SELECT must be INTEGER, CHARACTER or LOGICAL, is %s"
		, etype_name (typ->etype));
		fatal (serr);
	}

	if (rank) {
		sprintf (serr, "expression in SELECT must be scalar, is of rank %d"
		, rank);
		fatal (serr);
	}

/* block stuff */

	cac = make (CASE_CONSTRUCT);
	scac = &cac->CaseConstruct;
	block_push (cac);
	BLOCK_ADD_STMTS_TO (dummy_node);
	scac->SelectCaseStmt = root;
	scac->label = CURRENT_LABEL;
	scac->file = input_filename;
	scac->line = zzlineno;

	IGNORE_RETURN;
}

Node *check_CaseStmt (root)
Node *root;
{
	struct S_CaseConstruct *scac;
	Node *cac, *case_block, *ele, *temp;
	int has_sub, has_range;
	char *select_name, *case_name;

	if ((cac = block_match ("CASE", CASE_CONSTRUCT)) == NN)
		IGNORE_RETURN;

	scac = &cac->CaseConstruct;

/* error checking */

	se_after_select = 0;

	has_sub = (root->CaseStmt.CaseSelector->CaseSelector.SubscriptList != NN)? 1:0;

	if (has_sub) {
		Node *sublist = root->CaseStmt.CaseSelector->CaseSelector.SubscriptList;
		Etype *select_etype = scac->SelectCaseStmt->SelectCaseStmt.Expr->com.typ->etype;
		Node *from, *to;
		double value;
		Node *ele, *temp;
		struct S_Subscript *sub;

/*!!! maybe add subcript # to error messages */

	for (LIST_LOOP (ele, sublist, temp)) {
		sub = &ele->Subscript;

		if (sub->Colon1 != NN || sub->Expr2 != NN  /* catches (...:stride) */
		|| (sub->Colon != NN && sub->Expr == NN && sub->Expr1 == NN))  /* catches (:) */
			fatal ("wrong case-value-range-list in CASE statement");

		has_range = (sub->Colon != NN);
		from = sub->Expr;
		to   = sub->Expr1;

		if ( (from  != NN && !expr_is_constant (from, &value))
		||   (to    != NN && !expr_is_constant (to, &value)) )
			fatal ("expression for CASE value must be constant");

		if (has_range && select_etype->type == TYP_LOGICAL)
			fatal ("range cannot be used with a SELECT expression of type LOGICAL");

		if ((from != NN && etype_ne (from->com.typ->etype, select_etype))
		||  (to   != NN && etype_ne (to->com.typ->etype, select_etype)))
			fatal ("CASE value doesn't have the type of the SELECT expression");
	}
	}
	else {
		/* check that there is only one DEFAULT */

		for (LIST_LOOP (ele, scac->CaseStmtBlockList, temp))
			if (ele->CaseStmtBlock.CaseStmt->CaseStmt.CaseSelector->CaseSelector.Name
			!= NN) {
				fatal ("too many CASE DEFAULTs");
				break;
			}
	}

/* block stuff */

	case_block = make (CASE_STMT_BLOCK);
	BLOCK_ADD_STMTS_TO (case_block->CaseStmtBlock.CaseBlock);
	case_block->CaseStmtBlock.CaseStmt = root;

	scac->CaseStmtBlockList
	= list_add_beg (scac->CaseStmtBlockList , case_block , CASE_STMT_BLOCK_LIST);

/* checks construct-name stuff */

	select_name = (scac->SelectCaseStmt->SelectCaseStmt.Name1 != NN)
	? scac->SelectCaseStmt->SelectCaseStmt.Name->com.str : CN;
	case_name = (root->CaseStmt.Name != NN) ? root->CaseStmt.Name->com.str : CN;

	if (case_name != CN) {
		if (select_name == CN)
			fatal ("this CASE should not have a construct-name");
		else if (U_strcmp_nocase (select_name, case_name))
			fatal ("construct-name in CASE doesn't match that of the SELECT");
	}

	IGNORE_RETURN;
}

Node *check_EndSelectStmt (root)
Node *root;
{
	Node *cac;
	struct S_CaseConstruct *scac;
	char *select_name, *endselect_name;

	if ((cac = block_match ("END SELECT", CASE_CONSTRUCT)) == NN)
		IGNORE_RETURN;

/* block stuff */

	cac = block_pop ();
	scac = &cac->CaseConstruct;
	scac->EndSelectStmt = root;

	cac = trans_CaseConstruct (cac);

	block_add_stmt_with_label (cac, scac->label, scac->file, scac->line);
	scac->trans = TRN;

/* error checking */

	se_after_select = 0;

	if (scac->CaseStmtBlockList == NN)
		warning ("SELECT statement without CASEs");

/*!!! maybe also check for empty CASE blocks, and for only CASE is DEFAULT */

/* checks construct-name stuff */

	select_name = (scac->SelectCaseStmt->SelectCaseStmt.Name1 != NN)
	? scac->SelectCaseStmt->SelectCaseStmt.Name->com.str : CN;
	endselect_name = (root->EndSelectStmt.Name != NN)
	? root->EndSelectStmt.Name->com.str : CN;

	if (select_name != CN && endselect_name != CN) {
		if (U_strcmp_nocase (select_name, endselect_name))
			fatal ("construct-name in END SELECT doesn't match that of the SELECT");
	}
	else if (select_name != CN && endselect_name == CN) {
		fatal ("missing construct-name in END SELECT statement");
	}
	else if (select_name == CN && endselect_name != CN) {
		fatal ("this END SELECT should not have a construct-name");
	}

	IGNORE_RETURN;
}

case_end_write ()
{
	se_after_select = 0;
}

/******************************** T R A N S **************************************/

Node *trans_CaseConstruct (root)
Node *root;
{
	struct S_SelectCaseStmt *select
	= &root->CaseConstruct.SelectCaseStmt->SelectCaseStmt;

/* so that fun() just gets called once */

	if (tree_has (select->Expr, FUNCTION_REFERENCE))
		select->Expr = temp_replace (select->Expr);

	return root;
}

/******************************** W R I T E **************************************/

write_CaseConstruct (node)
Node *node;
{
	struct S_CaseConstruct *scac = &node->CaseConstruct;
	Node *ifc = make (IF_CONSTRUCT);
	struct S_IfConstruct *sifc = &ifc->IfConstruct;
	Node *ele, *temp, *list;
	struct S_CaseStmtBlock *case_block;
	Node *elseif_block, *expr1, *expr2, *expr, *sub, *sublist, *expr_list, *temp1;
	int first = 1;
	Node *case_index = make_DumExpr_1_PIN (scac->SelectCaseStmt->SelectCaseStmt.Expr);
	int case_index_type = case_index->com.typ->etype->type;
	int has_colon, subnum;

	if ((list = scac->CaseStmtBlockList) == NN)
		return;

	for (LIST_LOOP (ele, list, temp)) {
		case_block = &ele->CaseStmtBlock;

		if (case_block->CaseStmt->CaseStmt.CaseSelector->CaseSelector.Name != NN) {
			sifc->ElseStmt = case_block->CaseStmt;
			sifc->ElseBlock = case_block->CaseBlock;
			break;
		}

		sublist = case_block->CaseStmt->CaseStmt.CaseSelector->CaseSelector.SubscriptList;

	subnum = 0;
	for (LIST_LOOP (sub, sublist, temp1)) {
		subnum++;

		expr1 = (sub->Subscript.Expr != NN)
		? make_DumExpr_1_PIN (sub->Subscript.Expr) : NN;
		expr2 = (sub->Subscript.Expr1 != NN)
		? make_DumExpr_1_PIN (sub->Subscript.Expr1) : NN;

		has_colon = (sub->Subscript.Colon != NN) ? 1 : 0;

		if (has_colon) {
			if (expr1 != NN && expr2 != NN) {
				expr = make_DumExpr_write (
					  make_DumExpr_write_paren (expr1, &RelOpLe, case_index)
					, &AndOp
					, make_DumExpr_write_paren (case_index, &RelOpLe, expr2)
				);
			}
			else if (expr1 != NN)
				expr = make_DumExpr_write (expr1, &RelOpLe, case_index);
			else if (expr2 != NN)
				expr = make_DumExpr_write (case_index, &RelOpLe, expr2);
		}
		else {
			if (case_index_type == TYP_LOGICAL)
				expr = make_DumExpr_write (case_index, &EquivOpT, expr1);
			else
				expr = make_DumExpr_write (case_index, &RelOpEq, expr1);
		}

		switch (subnum) {

		case 1:
			expr_list = expr;
			break;

		case 2:
			expr_list = make_DumExpr_1_PIN (expr_list);
			expr      = make_DumExpr_1_PIN (expr);
			expr_list = make_DumExpr_write (expr_list, &OrOp, expr);
			break;

		default:
			expr      = make_DumExpr_1_PIN (expr);
			expr_list = make_DumExpr_write (expr_list, &OrOp, expr);
			break;
		}
	}

		if (first) {
			first = 0;
			sifc->IfThenStmt = make_IfThenStmt (expr_list);
			sifc->IfBlock = case_block->CaseBlock;
		}
		else {
			elseif_block = make (ELSE_IF_STMT_BLOCK);
			elseif_block->ElseIfStmtBlock.ElseIfStmt = make_ElseIfStmt (expr_list);
			elseif_block->ElseIfStmtBlock.ElseIfBlock = case_block->CaseBlock;

			sifc->ElseIfStmtBlockList = list_add_beg
			(sifc->ElseIfStmtBlockList, elseif_block, ELSE_IF_STMT_BLOCK_LIST);
		}
	}

	if (sifc->IfThenStmt == NN)  /* can happend if the only case is DEFAULT */
		sifc->IfThenStmt = make_IfThenStmt (make_Name (".TRUE."));

	write_IfConstruct (ifc);
}
